from django.contrib import admin

from jangle import models


@admin.register(models.ISOLanguageCodes)
class ISOLanguageCodesAdmin(admin.ModelAdmin):
    list_display = [
        "part_2b",
        "part_2t",
        "part_1",
    ]
    search_fields = [
        "part_1",
        "part_2b",
        "part_2t",
    ]


@admin.register(models.SimpleISOLanguageCollection)
class SimpleISOLanguageCollectionAdmin(admin.ModelAdmin):
    list_display = [
        "part_5",
        "names_en",
    ]
    search_fields = [
        "part_5",
        "names_en",
        "names_fr",
    ]


@admin.register(models.ISOLanguage)
class ISOLanguageAdmin(admin.ModelAdmin):
    list_display = [
        "ref_name",
        "part_3",
        "scope",
        "lang_type",
        "macrolanguage",
    ]
    search_fields = [
        "ref_name",
        "part_3",
        "codes__part_1",
        "codes__part_2b",
        "codes__part_2t",
    ]


@admin.register(models.ISOLanguageName)
class ISOLanguageNameAdmin(admin.ModelAdmin):
    list_display = ["printable", "iso_lang"]
    search_fields = [
        "printable",
        "inverted",
        "iso_lang__part_3",
        "iso_lang__codes__part_1",
        "iso_lang__codes__part_2b",
        "iso_lang__codes__part_2t",
    ]


@admin.register(models.Script)
class ScriptAdmin(admin.ModelAdmin):
    list_display = [
        "code",
        "no",
        "names_en",
        "pva",
        "unicode_version",
        "script_date",
    ]
    search_fields = [
        "code",
        "no",
        "names_en",
        "names_fr",
        "pva",
    ]


@admin.register(models.IANASubtagRegistry)
class IANASubtagRegistryAdmin(admin.ModelAdmin):
    list_display = [
        "file_date",
        "saved",
    ]


@admin.register(models.IANASubtagRecord)
class IANASubtagRecordAdmin(admin.ModelAdmin):
    list_display = [
        "first_description",
        "deprecated",
        "added",
        "pref_value",
    ]
    search_fields = [
        "descriptions__text",
    ]


@admin.register(models.LanguageSubtag)
class LanguageSubtagAdmin(admin.ModelAdmin):
    list_display = ["code", "iana", "scope", "macrolanguage"]
    search_fields = ["code", "iana__descriptions__text"]


@admin.register(models.ExtlangSubtag)
class ExtlangSubtagAdmin(admin.ModelAdmin):
    list_display = ["code", "iana", "scope", "macrolanguage"]
    search_fields = ["code", "iana__descriptions__text"]


@admin.register(models.ScriptSubtag)
class ScriptSubtagAdmin(admin.ModelAdmin):
    list_display = [
        "script",
        "iana",
    ]
    search_fields = [
        "script__code",
        "iana__descriptions__text",
    ]


@admin.register(models.RegionSubtag)
class RegionSubtagAdmin(admin.ModelAdmin):
    list_display = [
        "code",
        "region",
        "iana",
    ]
    search_fields = [
        "code",
        "iana__descriptions__text",
    ]


@admin.register(models.VariantSubtag)
class VariantSubtagAdmin(admin.ModelAdmin):
    list_display = [
        "text",
        "iana",
    ]
    search_fields = [
        "text",
        "iana__descriptions__text",
    ]


@admin.register(models.LanguageTag)
class LanguageTagAdmin(admin.ModelAdmin):
    list_display = [
        "tag_str",
        "description",
        "iana",
    ]
    search_fields = [
        "lang_tag__code",
        "region__code",
        "script__code",
        "variants__text",
        "private",
        "grandfathered_tag",
        "iana__descriptions__text",
    ]


@admin.register(models.LanguageTagVariantSubtag)
class LanguageTagVariantSubtag(admin.ModelAdmin):
    list_display = ["tag", "index", "variant"]
    search_fields = [
        "variant__text",
        "variant__iana__descriptions__text",
    ]


@admin.register(models.ExtensionSubtag)
class ExtensionSubtagAdmin(admin.ModelAdmin):
    list_display = [
        "tag",
        "index",
        "singleton",
    ]
    search_fields = [
        "texts__text",
    ]


@admin.register(models.ExtensionSubtagText)
class ExtensionSubtagTextAdmin(admin.ModelAdmin):
    list_display = [
        "extension",
        "index",
        "text",
    ]



from __future__ import annotations

import csv
import io
import os
import warnings
from typing import Generator, Iterable, Optional, TypeVar
from zipfile import ZipFile

import requests
from requests.compat import urljoin
from requests.models import ITER_CHUNK_SIZE

T = TypeVar("T")

IANA_ASSIGNMENTS_URL = "https://www.iana.org/assignments/"
SIL_ISO_639_DOWNLOADS_URL = (
    "https://iso639-3.sil.org/sites/iso639-3/files/downloads/"
)
SIL_ISO_639_LATEST = "20220311"
SIL_ISO_639_ZIPFILE = f"iso-639-3_Code_Tables_{SIL_ISO_639_LATEST}"


class Record(dict[str, list[str]]):
    """Used for working with record-jar records."""

    def add(self, key: str, val: str) -> None:
        """Adds a value to a field."""
        if key in self:
            self[key].append(val)
        else:
            self[key] = [val]

    def one(self, key: str) -> str:
        """Return a single value from a field.

        Raises
        ------
        ValueError
            If the field has multiple values.
        KeyError
            If the field has no values.
        """
        vals = self[key]
        if len(vals) > 1:
            raise ValueError(f"key '{key}' has multiple values {vals}")
        if not vals:
            raise KeyError(f"key '{key}' has an empty list of values")
        return vals[0]

    def get_one(self, key: str, default: T = None) -> str | T:
        """Return a single value from a field, or `default`.

        Raises
        ------
        ValueError
            If the field has multiple values.
        """
        try:
            return self.one(key)
        except KeyError:
            return default


def parse_record_jar(
    lines: Iterable[str], indent="\t", multiline_separator="\r\n"
) -> Generator[Record, None, None]:
    """Yields records from a set of lines.
    See https://datatracker.ietf.org/doc/pdf/draft-phillips-record-jar-02.
    """
    record = Record()
    key = None
    for line in lines:
        line_text = line.strip()
        if not line_text:
            continue
        if line.startswith(indent):
            if key is None:
                continue
            record[key][-1] += multiline_separator + line_text
        elif line_text == "%%":
            yield record
            record = Record()
        elif ":" in line:
            key, val = line_text.split(":", 1)
            record.add(key.strip(), val.strip())
    yield record


class SilTableReader:
    """Uses `csv.DictReader` to read tab-delimited data
    from https://iso639-3.sil.org/sites/iso639-3/files/downloads/,
    or a ZipFile to minimize requests.

    More information is available at
    https://iso639-3.sil.org/code_tables/download_tables.
    """

    chunk_size = ITER_CHUNK_SIZE

    def __init__(self, fn: str, zf: Optional[ZipFile] = None) -> None:
        self._f = None
        if zf:
            try:
                self._f = zf.open(
                    os.path.join(
                        SIL_ISO_639_ZIPFILE,
                        f"{fn}_{SIL_ISO_639_LATEST}.tab",
                    ),
                    "r",
                )
            except BaseException as exc:  # TODO
                warnings.warn(str(exc))
        if self._f is None:
            self._r = requests.get(
                urljoin(SIL_ISO_639_DOWNLOADS_URL, fn + ".tab"), stream=True
            )
            self._r.raise_for_status()
            self._r.encoding = "utf-8"

    def __iter__(self) -> csv.DictReader:
        if self._f:
            it = io.TextIOWrapper(self._f).readlines()
        else:
            it = self._r.iter_lines(self.chunk_size, decode_unicode=True)
        return csv.DictReader(it, dialect="excel-tab")

    def close(self) -> None:
        if self._f:
            self._f.close()
        elif self._r:
            self._r.close()

    def __enter__(self) -> SilTableReader:
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()


class IANARegistryReader:
    """Provides a file date and an iterable of records from an IANA registry."""

    chunk_size = ITER_CHUNK_SIZE
    records: Generator[Record, None, None]
    file_date: str

    def __init__(self, fn: str) -> None:
        response = requests.get(
            urljoin(IANA_ASSIGNMENTS_URL, "/".join([fn, fn])),
            stream=True,
        )
        response.raise_for_status()
        self.records = parse_record_jar(
            response.iter_lines(self.chunk_size, decode_unicode=True),
            indent=" " * 2,
            multiline_separator=" ",
        )
        self.file_date = next(self.records).one("File-Date")


class IANASubtagRegistryReader(IANARegistryReader):
    """Reads https://www.iana.org/assignments/language-subtag-registry/language-subtag-registry."""
    def __init__(self) -> None:
        super().__init__("language-subtag-registry")


class IANAExtensionsRegistryReader(IANARegistryReader):
    def __init__(self) -> None:
        super().__init__("language-extensions-registry")



from __future__ import annotations

import re
from dataclasses import dataclass, field
from typing import Optional

from jangle.patterns import RULES, match_rule
from jangle.utils import StrReprCls, split_subtags


@dataclass(repr=False)
class Extension(StrReprCls):
    singleton: str
    texts: list[str]

    @classmethod
    def from_match(cls, match: re.Match) -> Extension:
        return cls(
            singleton=match.group("singleton"),
            texts=split_subtags(match.group("ext_text")),
        )

    @classmethod
    def from_str(cls, string: str) -> Extension:
        return cls.from_match(match_rule("extension", string))

    def __str__(self) -> str:
        return "-".join([self.singleton, *self.texts])


@dataclass(repr=False)
class LangTag(StrReprCls):
    lang: str
    extlang: Optional[str] = None
    script: Optional[str] = None
    region: Optional[str] = None
    variants: list[str] = field(default_factory=list)
    extensions: list[Extension] = field(default_factory=list)
    private: Optional[str] = None

    @classmethod
    def from_groups(cls, groups: dict[str, str]) -> LangTag:
        langtag = cls(
            lang=groups["iso_639"].lower(),
            extlang=groups["extlang"].lower() or None,
            script=groups["script"].title() or None,
            region=groups["region"].upper() or None,
            private=groups["private_subtag"].lower().removeprefix("x-")
            or None,
        )
        if groups["variants"]:
            langtag.variants = split_subtags(groups["variants"].lower())
        langtag.extensions = list(
            map(
                Extension.from_match,
                RULES["extension"].finditer(groups["extensions"].lower()),
            )
        )
        return langtag

    @classmethod
    def from_match(cls, match: re.Match[str]) -> LangTag:
        return cls.from_groups(match.groupdict(""))

    @classmethod
    def from_str(cls, string: str) -> LangTag:
        return cls.from_match(match_rule("langtag", string))

    def __str__(self) -> str:
        subtags = [self.lang]
        subtags.extend(
            filter(
                None,
                [
                    self.extlang,
                    self.script,
                    self.region,
                ],
            )
        )
        subtags.extend(self.variants)
        subtags.extend(map(str, self.extensions))
        if self.private:
            subtags.extend(["x", self.private])
        return "-".join(subtags)

    def __contains__(self, val: LangTag | re.Match[str] | str) -> bool:
        if isinstance(val, re.Match):
            val = self.__class__.from_match(val)
        elif isinstance(val, str):
            val = self.__class__.from_str(val)
        if val.lang != self.lang:
            return False
        for attr_name in ["script", "extlang", "region", "private"]:
            attr = getattr(val, attr_name)
            if attr and attr != getattr(self, attr_name):
                return False
        for variant in val.variants:
            if variant not in self.variants:
                return False
        self_extensions = set(map(str, self.extensions))
        for extension in val.extensions:
            if str(extension) not in self_extensions:
                return False
        return True


@dataclass(repr=False)
class LanguageTag(StrReprCls):
    langtag: Optional[LangTag]
    private: Optional[str]
    grandfathered: Optional[str]

    @classmethod
    def from_match(cls, match: re.Match) -> LanguageTag:
        groups = match.groupdict("")
        lite = cls(
            langtag=None,
            private=groups["private_tag"].lower().removeprefix("x-") or None,
            grandfathered=groups["grandfathered"].lower() or None,
        )
        if groups["langtag"]:
            lite.langtag = LangTag.from_groups(groups)
        return lite

    @classmethod
    def from_str(cls, string: str) -> LanguageTag:
        return cls.from_match(match_rule("Language-Tag", string))

    def __str__(self) -> str:
        if self.grandfathered:
            return self.grandfathered
        if self.private:
            return "-".join(["x", self.private])
        return str(self.langtag)



# Generated by Django 4.1.3 on 2022-11-13 02:48

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="ExtensionSubtag",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("singleton", models.CharField(max_length=1)),
                ("index", models.PositiveSmallIntegerField(default=0)),
            ],
            options={
                "ordering": ["tag", "index"],
            },
        ),
        migrations.CreateModel(
            name="ExtlangSubtag",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("code", models.CharField(max_length=3, unique=True)),
                ("macrolanguage", models.CharField(max_length=3, null=True)),
                (
                    "scope",
                    models.CharField(
                        choices=[
                            ("I", "individual"),
                            ("C", "collection"),
                            ("M", "macrolanguage"),
                            ("S", "special"),
                        ],
                        default="I",
                        max_length=1,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="IANASubtagRecord",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("deprecated", models.DateField(null=True)),
                ("added", models.DateField()),
                ("comments", models.TextField(null=True)),
                (
                    "pref_value",
                    models.CharField(
                        max_length=42,
                        null=True,
                        verbose_name="preferred value",
                    ),
                ),
            ],
            options={
                "verbose_name": "IANA (sub)tag",
            },
        ),
        migrations.CreateModel(
            name="IANASubtagRegistry",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("file_date", models.DateField(unique=True)),
                ("saved", models.DateTimeField()),
            ],
            options={
                "verbose_name": "IANA Language Subtag Registry",
            },
        ),
        migrations.CreateModel(
            name="ISOLanguage",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "ref_name",
                    models.CharField(
                        max_length=150,
                        unique=True,
                        verbose_name="reference name",
                    ),
                ),
                (
                    "part_3",
                    models.CharField(
                        max_length=3,
                        unique=True,
                        verbose_name="alpha-3 code (ISO 639-3)",
                    ),
                ),
                (
                    "lang_type",
                    models.CharField(
                        choices=[
                            ("A", "ancient"),
                            ("C", "constructed"),
                            ("E", "extinct"),
                            ("H", "historical"),
                            ("L", "living"),
                            ("S", "special"),
                        ],
                        max_length=1,
                        verbose_name="type",
                    ),
                ),
                (
                    "scope",
                    models.CharField(
                        choices=[
                            ("I", "individual"),
                            ("M", "macrolanguage"),
                            ("S", "special"),
                        ],
                        max_length=1,
                    ),
                ),
                ("comment", models.CharField(max_length=150, null=True)),
            ],
            options={
                "verbose_name": "ISO 639-3 language",
            },
        ),
        migrations.CreateModel(
            name="ISOLanguageCodes",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "part_2b",
                    models.CharField(
                        max_length=3,
                        unique=True,
                        verbose_name="bibliographic code (ISO 639-2/B)",
                    ),
                ),
                (
                    "part_2t",
                    models.CharField(
                        max_length=3,
                        unique=True,
                        verbose_name="terminological code (ISO 639-2/T)",
                    ),
                ),
                (
                    "part_1",
                    models.CharField(
                        max_length=2,
                        null=True,
                        verbose_name="alpha-2 code (ISO 639-1)",
                    ),
                ),
                (
                    "names_en",
                    models.CharField(
                        max_length=75,
                        unique=True,
                        verbose_name="noms français",
                    ),
                ),
                (
                    "names_fr",
                    models.CharField(
                        max_length=75,
                        unique=True,
                        verbose_name="English names",
                    ),
                ),
            ],
            options={
                "verbose_name": "ISO 639-2 and 639-1 language codeset",
            },
        ),
        migrations.CreateModel(
            name="LanguageSubtag",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("code", models.CharField(max_length=3, unique=True)),
                ("macrolanguage", models.CharField(max_length=3, null=True)),
                (
                    "scope",
                    models.CharField(
                        choices=[
                            ("I", "individual"),
                            ("C", "collection"),
                            ("M", "macrolanguage"),
                            ("S", "special"),
                        ],
                        default="I",
                        max_length=1,
                    ),
                ),
                (
                    "iana",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="jangle.ianasubtagrecord",
                    ),
                ),
                (
                    "iso_lang",
                    models.OneToOneField(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="subtag",
                        to="jangle.isolanguage",
                    ),
                ),
                (
                    "iso_lang_codes",
                    models.OneToOneField(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="subtag",
                        to="jangle.isolanguagecodes",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="LanguageTag",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "grandfathered_tag",
                    models.CharField(max_length=42, null=True),
                ),
                ("private", models.CharField(max_length=42, null=True)),
                (
                    "extlang",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="jangle.extlangsubtag",
                    ),
                ),
                (
                    "iana",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="jangle.ianasubtagrecord",
                    ),
                ),
                (
                    "lang",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="jangle.languagesubtag",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Region",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "tier",
                    models.PositiveSmallIntegerField(
                        choices=[
                            (0, "global"),
                            (1, "regional"),
                            (2, "continental"),
                            (3, "sub-regional"),
                            (4, "intermediary"),
                            (5, "country or area"),
                        ]
                    ),
                ),
                (
                    "no",
                    models.PositiveSmallIntegerField(
                        unique=True,
                        verbose_name="ISO 3166-1 numeric / UN M.49 code",
                    ),
                ),
                (
                    "parent",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="jangle.region",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Script",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "code",
                    models.CharField(
                        max_length=4,
                        unique=True,
                        verbose_name="ISO 15924 code",
                    ),
                ),
                (
                    "no",
                    models.PositiveSmallIntegerField(
                        unique=True, verbose_name="ISO 15924 number"
                    ),
                ),
                (
                    "names_en",
                    models.CharField(
                        max_length=75,
                        unique=True,
                        verbose_name="English names",
                    ),
                ),
                (
                    "names_fr",
                    models.CharField(
                        max_length=75,
                        unique=True,
                        verbose_name="noms français",
                    ),
                ),
                (
                    "pva",
                    models.CharField(
                        max_length=150,
                        null=True,
                        verbose_name="property value alias",
                    ),
                ),
                (
                    "unicode_version",
                    models.CharField(max_length=12, null=True),
                ),
                ("script_date", models.DateField(verbose_name="date")),
            ],
            options={
                "verbose_name": "ISO 15924 script",
            },
        ),
        migrations.CreateModel(
            name="SimpleISOLanguageCollection",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "part_5",
                    models.CharField(
                        max_length=3,
                        unique=True,
                        verbose_name="ISO 639-5 alpha-3 code",
                    ),
                ),
                (
                    "names_en",
                    models.CharField(
                        max_length=75,
                        unique=True,
                        verbose_name="English names",
                    ),
                ),
                (
                    "names_fr",
                    models.CharField(
                        max_length=75,
                        unique=True,
                        verbose_name="noms français",
                    ),
                ),
            ],
            options={
                "verbose_name": "simple ISO 639-5 language collection",
            },
        ),
        migrations.CreateModel(
            name="VariantSubtag",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("text", models.CharField(max_length=8, unique=True)),
                (
                    "iana",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="jangle.ianasubtagrecord",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="ScriptSubtag",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "iana",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="jangle.ianasubtagrecord",
                    ),
                ),
                (
                    "script",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="subtag",
                        to="jangle.script",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="RegionSubtag",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("code", models.CharField(max_length=3, unique=True)),
                (
                    "ext_data",
                    models.OneToOneField(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="subtag",
                        to="jangle.region",
                    ),
                ),
                (
                    "iana",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="jangle.ianasubtagrecord",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="LanguageTagVariantSubtag",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("index", models.PositiveSmallIntegerField(default=0)),
                (
                    "tag",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="variants_through",
                        to="jangle.languagetag",
                    ),
                ),
                (
                    "variant",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="tags_through",
                        to="jangle.variantsubtag",
                    ),
                ),
            ],
            options={
                "ordering": ["tag", "index"],
                "unique_together": {("tag", "variant"), ("tag", "index")},
            },
        ),
        migrations.AddField(
            model_name="languagetag",
            name="region",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="jangle.regionsubtag",
            ),
        ),
        migrations.AddField(
            model_name="languagetag",
            name="script",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="jangle.scriptsubtag",
            ),
        ),
        migrations.AddField(
            model_name="languagetag",
            name="variants",
            field=models.ManyToManyField(
                through="jangle.LanguageTagVariantSubtag",
                to="jangle.variantsubtag",
            ),
        ),
        migrations.AddField(
            model_name="languagesubtag",
            name="iso_lang_collection",
            field=models.OneToOneField(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="subtag",
                to="jangle.simpleisolanguagecollection",
            ),
        ),
        migrations.AddField(
            model_name="languagesubtag",
            name="suppress_script",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="jangle.script",
            ),
        ),
        migrations.CreateModel(
            name="ISORegion",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "alpha_2",
                    models.CharField(
                        max_length=2,
                        unique=True,
                        verbose_name="ISO 3166-1 alpha-2 code",
                    ),
                ),
                (
                    "alpha_3",
                    models.CharField(
                        max_length=3,
                        unique=True,
                        verbose_name="ISO 3166-1 alpha-3 code",
                    ),
                ),
                (
                    "in_ldc",
                    models.BooleanField(
                        default=False, verbose_name="Least Developed Countries"
                    ),
                ),
                (
                    "in_lldc",
                    models.BooleanField(
                        default=False,
                        verbose_name="Land Locked Developing Countries",
                    ),
                ),
                (
                    "in_sids",
                    models.BooleanField(
                        default=False,
                        verbose_name="Small Island Developing States",
                    ),
                ),
                (
                    "region",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="iso",
                        to="jangle.region",
                    ),
                ),
            ],
            options={
                "verbose_name": "ISO 3166 Region",
            },
        ),
        migrations.CreateModel(
            name="ISOLanguageName",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "printable",
                    models.CharField(
                        max_length=75, verbose_name="printable translated name"
                    ),
                ),
                (
                    "inverted",
                    models.CharField(
                        max_length=75, verbose_name="inverted translated name"
                    ),
                ),
                (
                    "iso_lang",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="names",
                        to="jangle.isolanguage",
                    ),
                ),
            ],
            options={
                "verbose_name": "ISO 639-3 language name",
            },
        ),
        migrations.AddField(
            model_name="isolanguage",
            name="codes",
            field=models.OneToOneField(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="language",
                to="jangle.isolanguagecodes",
            ),
        ),
        migrations.AddField(
            model_name="isolanguage",
            name="macrolanguage",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="individuals",
                to="jangle.isolanguage",
            ),
        ),
        migrations.AddField(
            model_name="ianasubtagrecord",
            name="registry",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="records",
                to="jangle.ianasubtagregistry",
            ),
        ),
        migrations.CreateModel(
            name="IANASubtagDescription",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("text", models.CharField(max_length=75)),
                ("index", models.PositiveSmallIntegerField(default=0)),
                (
                    "subtag",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="descriptions",
                        to="jangle.ianasubtagrecord",
                    ),
                ),
            ],
            options={
                "ordering": ["subtag", "index"],
            },
        ),
        migrations.AddField(
            model_name="extlangsubtag",
            name="iana",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="jangle.ianasubtagrecord",
            ),
        ),
        migrations.AddField(
            model_name="extlangsubtag",
            name="iso_lang",
            field=models.OneToOneField(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="ext_subtag",
                to="jangle.isolanguage",
            ),
        ),
        migrations.AddField(
            model_name="extlangsubtag",
            name="suppress_script",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="jangle.script",
            ),
        ),
        migrations.CreateModel(
            name="ExtensionSubtagText",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("text", models.CharField(max_length=8)),
                ("index", models.PositiveSmallIntegerField(default=0)),
                (
                    "extension",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="texts",
                        to="jangle.extensionsubtag",
                    ),
                ),
            ],
            options={
                "ordering": ["extension", "index"],
            },
        ),
        migrations.AddField(
            model_name="extensionsubtag",
            name="tag",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="extensions",
                to="jangle.languagetag",
            ),
        ),
        migrations.CreateModel(
            name="RegionName",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField(max_length=150)),
                (
                    "iso_lang",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="jangle.isolanguage",
                    ),
                ),
                (
                    "region",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="names",
                        to="jangle.region",
                    ),
                ),
            ],
            options={
                "unique_together": {("region", "iso_lang")},
            },
        ),
        migrations.AddConstraint(
            model_name="languagetag",
            constraint=models.CheckConstraint(
                check=models.Q(
                    ("lang__isnull", False),
                    ("grandfathered_tag__isnull", False),
                    ("private__isnull", False),
                    _connector="OR",
                ),
                name="langtag_or_grandfathered_or_private",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="ianasubtagdescription",
            unique_together={("subtag", "text"), ("subtag", "index")},
        ),
        migrations.AlterUniqueTogether(
            name="extensionsubtagtext",
            unique_together={("extension", "index")},
        ),
        migrations.AlterUniqueTogether(
            name="extensionsubtag",
            unique_together={("tag", "singleton"), ("tag", "index")},
        ),
    ]



import re
import tempfile

_SUBTAG_LOOKAHEAD = r"(?![^\-\s])"
"""Checks if there is no subtag immediately ahead."""

_ALPHANUM = r"[A-Za-z\d]"
_REGULAR = r"|".join(
    map(
        re.escape,
        [
            "art-lojban",
            "cel-gaulish",
            "no-bok",
            "no-nyn",
            "zh-guoyo",
            "zh-hakka",
            "zh-min",
            "zh-min-nan",
            "zh-xiang",
        ],
    )
)
_IRREGULAR = (
    r"en\-GB\-oed|i\-(?:%s)|sgn\-(?:BE\-FR|BE\-NL|CH\-DE)"
    % r"|".join(
        [
            "ami",
            "bnn",
            "default",
            "enochian",
            "hak",
            "klingon",
            "lux",
            "mingo",
            "navajo",
            "tao",
            "tay",
            "tsu",
        ]
    )
)
_GRANDFATHERED = r"(?P<regular>%s)|(?P<irregular>%s)" % (_REGULAR, _IRREGULAR)
_SINGLETON = r"[A-WY-Za-wy-z\d]"
_PRIVATEUSE = r"x(?:\-%s{1,8})+" % _ALPHANUM
_EXTENSION = r"(?P<singleton>%s)(?P<ext_text>(?:\-%s{2,8})+)" % (
    _SINGLETON,
    _ALPHANUM,
)
_VARIANT = r"%s{5,8}|\d%s{3}" % (_ALPHANUM, _ALPHANUM)
_REGION = r"(?P<iso_3166>[A-Za-z]{2})|(?P<un_m49>\d{3})"
_SCRIPT = r"[A-Za-z]{4}"
_EXTLANG = r"(?P<extlang_iso_639>[A-Za-z]{3})(?P<extlang_reserved>(?:\-[A-Za-z]{3}){0,2})"
_LANGUAGE = r"(?P<iso_639>[A-Za-z]{2,3})(?:\-(?P<extlang>%s))?" % _EXTLANG
_LANGTAG = (
    r"%s%s(?:\-(?P<script>%s)%s)?(?:\-(?P<region>%s)%s)?(?P<variants>(?:\-(?:%s))*)(?P<extensions>(?:\-%s)*)(?:\-(?P<private_subtag>%s))?"
    % (
        _LANGUAGE,
        _SUBTAG_LOOKAHEAD,
        _SCRIPT,
        _SUBTAG_LOOKAHEAD,
        _REGION,
        _SUBTAG_LOOKAHEAD,
        _VARIANT,
        _EXTENSION,
        _PRIVATEUSE,
    )
)

RULES: dict[str, re.Pattern[str]] = {
    key: re.compile(pattern, flags=re.I)
    for key, pattern in {
        "Language-Tag": r"(?P<grandfathered>%s)|(?P<private_tag>%s)|(?P<langtag>%s)"
        % (_GRANDFATHERED, _PRIVATEUSE, _LANGTAG),
        "langtag": _LANGTAG,
        "language": _LANGUAGE,
        "extlang": _EXTLANG,
        "script": _SCRIPT,
        "region": _REGION,
        "variant": _VARIANT,
        "extension": _EXTENSION,
        "singleton": _SINGLETON,
        "privateuse": _PRIVATEUSE,
        "grandfathered": _GRANDFATHERED,
        "regular": _REGULAR,
        "irregular": _IRREGULAR,
        "alphanum": _ALPHANUM,
    }.items()
}
"""RegEx patterns for rules from the
RFC 5646 ABNF syntax definition.
See https://www.rfc-editor.org/rfc/rfc5646.html#section-2.1.
"""


def match_rule(rule: str, string: str) -> re.Match[str]:
    match = RULES[rule].fullmatch(string)
    if match is None:
        raise ValueError(f"invalid {rule}: {string}")
    return match


def rules_rst() -> str:
    """Saves a temporary reStructuredText file
    documenting all rule patterns, used in `docs/source/patterns.rst`
    """
    _, temp = tempfile.mkstemp(
        text=True,
        prefix="jangle_rfc5646_rules_",
        suffix=".rst",
    )
    with open(temp, "w") as f:
        for rule, pattern in RULES.items():
            f.write(f"{rule}:\r\n\r\n")
            f.write(
                ":regexp:`%s`\r\n\r\n" % pattern.pattern.replace("\\", r"\\")
            )
    return temp



from .languages import (
    ISOLanguage,
    ISOLanguageCodes,
    ISOLanguageName,
    SimpleISOLanguageCollection,
)
from .regions import ISORegion, Region, RegionName
from .scripts import Script
from .tags import (
    ExtensionSubtag,
    ExtensionSubtagText,
    ExtlangSubtag,
    IANASubtagDescription,
    IANASubtagRecord,
    IANASubtagRegistry,
    LanguageSubtag,
    LanguageTag,
    LanguageTagVariantSubtag,
    RegionSubtag,
    ScriptSubtag,
    VariantSubtag,
)

__all__ = [
    "SimpleISOLanguageCollection",
    "ISOLanguageCodes",
    "ISOLanguage",
    "ISOLanguageName",
    "Script",
    "Region",
    "RegionName",
    "ISORegion",
    "IANASubtagRegistry",
    "IANASubtagRecord",
    "IANASubtagDescription",
    "LanguageSubtag",
    "ExtlangSubtag",
    "ScriptSubtag",
    "RegionSubtag",
    "VariantSubtag",
    "LanguageTag",
    "LanguageTagVariantSubtag",
    "ExtensionSubtag",
    "ExtensionSubtagText",
]



from __future__ import annotations

import operator
import re
import warnings
from datetime import datetime
from functools import cached_property, reduce
from typing import Any, Tuple

from django.conf import settings
from django.db import models
from django.utils.timezone import utc

import jangle.lite
from jangle.readers import IANASubtagRegistryReader
from jangle.utils import choice_from_iana

from .languages import (
    ISOLanguage,
    ISOLanguageCodes,
    SimpleISOLanguageCollection,
)
from .regions import Region
from .scripts import Script


class IANASubtagRegistryManager(models.Manager["IANASubtagRegistry"]):
    def register(self, clear=True, descriptions_batch_size=64) -> None:
        """Saves language, extlang, script, region, grandfathered, and redundant
        language tags and subtags from the IANA language subtag registry
        to their corresponding tables.
        """
        if clear:
            self.all().delete()
        registry = IANASubtagRegistryReader()
        registry_obj = self.create(
            file_date=registry.file_date,
            saved=datetime.utcnow().replace(tzinfo=utc),
        )
        descriptions = []
        tag_strs = []
        for i, record in enumerate(registry.records):
            if i and not i % descriptions_batch_size:
                IANASubtagDescription.objects.bulk_create(descriptions)
                descriptions = []
            if "Subtag" in record and ".." in record.one("Subtag"):
                continue  # private use
            iana = IANASubtagRecord.objects.create(
                registry=registry_obj,
                deprecated=record.get_one("Deprecated"),
                added=record.one("Added"),
                comments=record.get_one("Comments"),
                pref_value=record.get_one("Preferred-Value"),
            )
            descriptions.extend(
                IANASubtagDescription(
                    subtag=iana,
                    text=text,
                    index=i,
                )
                for i, text in enumerate(record["Description"])
            )
            record_type = record.one("Type")
            if record_type in {"language", "extlang"}:
                if record_type == "language":
                    subtag = LanguageSubtag()
                else:
                    subtag = ExtlangSubtag()
                subtag.iana = iana
                subtag.code = record.one("Subtag")
                subtag.macrolanguage = record.get_one("Macrolanguage")
                if "Scope" in record:
                    subtag.scope = choice_from_iana(
                        LangSubtagFromIANARecord.Scope, record.one("Scope")
                    )
                if "Suppress-Script" in record:
                    subtag.suppress_script = Script.objects.get(
                        code=record.one("Suppress-Script")
                    )
                subtag.save()
                if record_type == "language":
                    LanguageTag.objects.create(lang=subtag)
            elif record_type == "script":
                ScriptSubtag.objects.create(
                    iana=iana,
                    script=Script.objects.get(code=record.one("Subtag")),
                )
            elif record_type == "region":
                RegionSubtag.objects.create(
                    iana=iana, code=record.one("Subtag")
                )
            elif record_type == "variant":
                VariantSubtag.objects.create(
                    iana=iana, text=record.one("Subtag")
                )
            elif record_type == "grandfathered":
                for tag_str, tag_iana in tag_strs:
                    LanguageTag.objects.create_from_langtag(
                        jangle.lite.LangTag.from_str(tag_str),
                        bool(tag_iana.deprecated),
                        iana=tag_iana,
                    )
                tag_strs = []
                tag = LanguageTag.objects.create(
                    iana=iana,
                    grandfathered_tag=record.one("Tag"),
                )
                tag.save()
            elif record_type == "redundant":
                tag, created = LanguageTag.objects.get_or_create_from_str(
                    record.one("Tag"),
                    allow_deprecated=bool(iana.deprecated),
                    defaults={
                        "iana": iana,
                    },
                )
                if not created:
                    tag.iana = iana
                    tag.save()
            for prefix in record.get("Prefix", []):
                tag_strs.append(
                    (
                        "-".join([prefix, record.one("Subtag")]),
                        iana,
                    )
                )
        IANASubtagDescription.objects.bulk_create(
            descriptions,
            batch_size=descriptions_batch_size,
        )


class IANASubtagRegistry(models.Model):
    """Represents a saved instance
    of the IANA Language Subtag Registry.

    See https://www.iana.org/assignments/iso_lang-subtags-templates/iso_lang-subtags-templates.xhtml,
    https://www.iana.org/assignments/language-subtag-registry/language-subtag-registry.
    """

    records: "models.manager.RelatedManager[IANASubtagRecord]"
    """"""

    file_date = models.DateField(unique=True)
    """"""
    saved = models.DateTimeField()
    """Time registry was saved."""

    def __str__(self) -> str:
        return f"{self.file_date} / {self.saved}"

    objects = IANASubtagRegistryManager()

    class Meta:
        verbose_name = "IANA Language Subtag Registry"


class IANASubtagRecord(models.Model):
    """Abstract model to represent a record in
    the IANA language subtag registry.
    """

    descriptions: "models.manager.RelatedManager[IANASubtagDescription]"
    """"""
    registry = models.ForeignKey(
        IANASubtagRegistry,
        related_name="records",
        on_delete=models.CASCADE,
    )
    """"""
    deprecated = models.DateField(null=True)
    """Date deprecated."""
    added = models.DateField()
    """Date added."""
    comments = models.TextField(null=True)
    """"""
    pref_value = models.CharField("preferred value", null=True, max_length=42)
    """Preferred value."""

    def first_description(self) -> str:
        return self.descriptions.get(index=0).text

    def __str__(self) -> str:
        return self.first_description()

    class Meta:
        verbose_name = "IANA (sub)tag"


class IANASubtagDescription(models.Model):
    subtag = models.ForeignKey(
        IANASubtagRecord,
        related_name="descriptions",
        on_delete=models.CASCADE,
    )
    """IANA subtag record."""
    text = models.CharField(max_length=75)
    """"""
    index = models.PositiveSmallIntegerField(default=0)
    """"""

    def __str__(self) -> str:
        return self.text

    class Meta:
        unique_together = (("subtag", "index"), ("subtag", "text"))
        ordering = ["subtag", "index"]


class SubtagFromIANARecord(models.Model):
    iana = models.OneToOneField(
        IANASubtagRecord,
        related_name="+",
        on_delete=models.CASCADE,
    )
    """IANA (sub)tag record."""

    class Meta:
        abstract = True


class LangSubtagFromIANARecord(SubtagFromIANARecord):
    class Scope(models.TextChoices):
        INDIVIDUAL = "I", "individual"
        COLLECTION = "C", "collection"
        MACROLANGUAGE = "M", "macrolanguage"
        SPECIAL = "S", "special"
        # PRIVATE_USE = 'P', 'private use'

    code = models.CharField(unique=True, max_length=3)
    """ISO 639 or registered code."""
    macrolanguage = models.CharField(max_length=3, null=True)
    """"""
    scope = models.CharField(
        choices=Scope.choices,
        default=Scope.INDIVIDUAL,
        max_length=1,
    )
    """Classification."""
    suppress_script = models.ForeignKey(
        Script,
        null=True,
        on_delete=models.SET_NULL,
    )
    """Script used to write
    the overwhelming majority of
    documents in this language.
    """

    def __str__(self) -> str:
        return self.code

    class Meta:
        abstract = True


class LanguageSubtag(LangSubtagFromIANARecord):
    iso_lang_codes = models.OneToOneField(
        ISOLanguageCodes,
        related_name="subtag",
        null=True,
        on_delete=models.CASCADE,
    )
    """"""
    iso_lang = models.OneToOneField(
        ISOLanguage,
        related_name="subtag",
        null=True,
        on_delete=models.CASCADE,
    )
    """"""
    iso_lang_collection = models.OneToOneField(
        SimpleISOLanguageCollection,
        related_name="subtag",
        null=True,
        on_delete=models.CASCADE,
    )
    """"""

    def save(self, *args, **kwargs) -> None:
        if self.scope == self.Scope.COLLECTION:
            if self.iso_lang_collection and self.iso_lang_codes is None:
                try:
                    self.iso_lang_collection = (
                        SimpleISOLanguageCollection.objects.get_from_ietf(
                            self.code
                        )
                    )
                except SimpleISOLanguageCollection.DoesNotExist:
                    self.iso_lang_codes = (
                        ISOLanguageCodes.objects.get_from_ietf(self.code)
                    )
        else:
            if self.iso_lang is None:
                try:
                    self.iso_lang = ISOLanguage.objects.get_from_ietf(
                        self.code
                    )
                except ISOLanguage.DoesNotExist:
                    warnings.warn(
                        f"could not find any ISO-639 codes from '{self.code}'"
                    )
        super().save(*args, **kwargs)


class ExtlangSubtag(LangSubtagFromIANARecord):
    iso_lang = models.OneToOneField(
        ISOLanguage,
        related_name="ext_subtag",
        null=True,
        on_delete=models.CASCADE,
    )
    """"""

    def save(self, *args, **kwargs) -> None:
        if not self.iso_lang:
            try:
                self.iso_lang = ISOLanguage.objects.get(part_3=self.code)
            except ISOLanguage.DoesNotExist:
                warnings.warn(
                    f"could not find any ISO-639 codes from '{self.code}'"
                )

        super().save(*args, **kwargs)


class ScriptSubtag(SubtagFromIANARecord):
    """"""

    script = models.OneToOneField(
        Script,
        related_name="subtag",
        on_delete=models.CASCADE,
    )
    """External ISO-15924 data from unicode.org."""

    def __str__(self) -> str:
        return self.script.code


class RegionSubtag(SubtagFromIANARecord):
    """"""

    code = models.CharField(unique=True, max_length=3)
    """"""
    region = models.OneToOneField(
        Region,
        null=True,
        related_name="subtag",
        on_delete=models.SET_NULL,
    )
    """External UN M.49 and ISO 3166 data."""

    def save(self, *args, **kwargs) -> None:
        if self.region is None:
            try:
                if len(self.code) == 2:
                    self.region = Region.objects.get(iso__alpha_2=self.code)
                else:
                    self.region = Region.objects.get(no=int(self.code))
            except Region.DoesNotExist:
                warnings.warn(f"could not find region of code '{self.code}'")
        super().save(*args, **kwargs)

    def __str__(self) -> str:
        return self.code


class VariantSubtag(SubtagFromIANARecord):
    """"""

    text = models.CharField(unique=True, max_length=8)
    """"""

    def __str__(self) -> str:
        return self.text


class LanguageTagQuerySet(models.QuerySet["LanguageTag"]):
    def active(self) -> LanguageTagQuerySet:
        """Excludes all deprecated tags and tags with
        deprecated subtags, according to the IANA subtag registry.
        """
        return self.exclude(
            iana__deprecated__isnull=False,
            **{
                f"{subtag}__iana__deprecated__isnull": False
                for subtag in ["extlang", "script", "region", "variants"]
            },
        )

    def private(self) -> LanguageTagQuerySet:
        """Private tags (not langtags with a private subtag)."""
        return self.filter(lang__isnull=True, private__isnull=False)

    def get_from_lite(self, lite: jangle.lite.LanguageTag) -> LanguageTag:
        if lite.private:
            return self.get(lang__isnull=True, private=lite.private)
        if lite.grandfathered:
            return self.get(grandfathered_tag=lite.grandfathered)
        queryset = self
        if lite.langtag is None:
            raise ValueError("empty language tag")
        if lite.langtag.variants:
            tag_variants_qs = (
                models.Q(variants__text=variant)
                for variant in lite.langtag.variants
            )
            queryset = queryset.filter(reduce(operator.and_, tag_variants_qs))
        queryset = queryset.exclude(
            variants_through__index=len(lite.langtag.variants)
        )
        if lite.langtag.extensions:
            extension_qs = (
                models.Q(
                    texts=reduce(
                        operator.and_,
                        (
                            models.Q(index=i, text__iexact=text)
                            for i, text in enumerate(val.texts)
                        ),
                    ),
                    index=i,
                    singleton__iexact=val.singleton,
                )
                for i, val in enumerate(lite.langtag.extensions)
            )
            queryset = queryset.filter(
                extensions=reduce(operator.and_, extension_qs)
            )
        queryset = queryset.exclude(
            extensions__index=len(lite.langtag.extensions)
        )
        return queryset.get(
            lang__code=lite.langtag.lang,
            extlang__code=lite.langtag.extlang,
            script__script__code=lite.langtag.script,
            region__code=lite.langtag.region,
            private=lite.langtag.private,
        )

    def get_from_match(self, match: re.Match[str]) -> LanguageTag:
        return self.get_from_lite(jangle.lite.LanguageTag.from_match(match))

    def get_from_str(self, string: str) -> LanguageTag:
        return self.get_from_lite(jangle.lite.LanguageTag.from_str(string))


class LanguageTagManager(models.Manager["LanguageTag"]):
    """"""

    def get_queryset(self) -> LanguageTagQuerySet:
        return LanguageTagQuerySet(self.model, using=self._db)

    def private(self) -> LanguageTagQuerySet:
        return self.get_queryset().private()

    def active(self) -> LanguageTagQuerySet:
        return self.get_queryset().active()

    def get_from_lite(self, lite: jangle.lite.LanguageTag) -> LanguageTag:
        return self.get_queryset().get_from_lite(lite)

    def get_from_match(self, match: re.Match[str]) -> LanguageTag:
        return self.get_queryset().get_from_match(match)

    def get_from_str(self, string: str) -> LanguageTag:
        return self.get_queryset().get_from_str(string)

    def create_from_langtag(
        self, langtag: jangle.lite.LangTag, allow_deprecated=False, **kwargs
    ) -> LanguageTag:
        subtag_kwargs = dict()
        if not allow_deprecated:
            subtag_kwargs["iana__deprecated__isnull"] = True
        tag = self.model(**kwargs)
        tag.lang = LanguageSubtag.objects.get(
            code=langtag.lang,
            **subtag_kwargs,
        )
        if langtag.extlang:
            tag.extlang = ExtlangSubtag.objects.get(
                code=langtag.extlang,
                **subtag_kwargs,
            )
        if langtag.script:
            tag.script = ScriptSubtag.objects.get(
                script__code=langtag.script,
                **subtag_kwargs,
            )
        if langtag.region:
            tag.region = RegionSubtag.objects.get(
                code=langtag.region,
                **subtag_kwargs,
            )
        tag.save()
        for i, variant in enumerate(langtag.variants):
            LanguageTagVariantSubtag.objects.create(
                tag=tag,
                variant=VariantSubtag.objects.get(text=variant),
                index=i,
            )
        for i, extension in enumerate(langtag.extensions):
            ext_obj = ExtensionSubtag.objects.create(
                tag=tag,
                singleton=extension.singleton,
                index=i,
            )
            for i, text in extension.texts:
                ExtensionSubtagText.objects.create(
                    extension=ext_obj,
                    text=text,
                    index=i,
                )
        return tag

    def get_or_create_from_str(
        self,
        string: str,
        allow_deprecated=False,
        defaults: dict[str, Any] = dict(),
    ) -> Tuple[LanguageTag, bool]:
        lite = jangle.lite.LanguageTag.from_str(string)
        queryset = self.get_queryset()
        if not allow_deprecated:
            queryset = queryset.active()
        try:
            return queryset.get_from_lite(lite), False
        except self.model.DoesNotExist:
            if lite.langtag:
                return (
                    self.create_from_langtag(
                        lite.langtag, allow_deprecated, **defaults
                    ),
                    True,
                )
            elif lite.private:
                return self.create(private=lite.private), True
            else:
                raise ValueError(f"cannot create grandfathered tag '{string}'")

    def native(self) -> LanguageTag:
        """Returns a LanguageTag from `settings.LANGUAGE_CODE`"""
        return self.get_from_str(settings.LANGUAGE_CODE)


class LanguageTag(models.Model):
    """Represents an RFC5646 language tag.

    See https://www.rfc-editor.org/rfc/rfc5646.html.
    """

    variants_through: "models.manager.RelatedManager[LanguageTagVariantSubtag]"
    """"""
    extensions: "models.manager.RelatedManager[ExtensionSubtag]"
    """"""
    iana = models.ForeignKey(
        IANASubtagRecord,
        null=True,
        on_delete=models.CASCADE,
    )
    """Original IANA (sub)tag record."""
    grandfathered_tag = models.CharField(null=True, max_length=42)
    """"""
    lang = models.ForeignKey(
        LanguageSubtag,
        null=True,
        on_delete=models.PROTECT,
    )
    """Language subtag."""
    extlang = models.ForeignKey(
        ExtlangSubtag,
        null=True,
        on_delete=models.PROTECT,
    )
    """Extlang subtag."""
    script = models.ForeignKey(
        ScriptSubtag,
        null=True,
        on_delete=models.PROTECT,
    )
    """Script subtag."""
    region = models.ForeignKey(
        RegionSubtag,
        null=True,
        on_delete=models.PROTECT,
    )
    """Region subtag."""
    variants = models.ManyToManyField(
        VariantSubtag,
        through="LanguageTagVariantSubtag",
        through_fields=("tag", "variant"),
    )
    private = models.CharField(
        null=True,
        max_length=42,
    )
    """Private tag or subtag."""

    @property
    def is_private(self) -> bool:
        return self.lang is None and self.private is not None

    @cached_property
    def pref_tag(self) -> LanguageTag:
        """Preferred tag as defined in the IANA registry, or `self`."""
        if not (self.iana and self.iana.pref_value):
            return self
        return self.__class__.objects.get_from_str(self.iana.pref_value)

    @cached_property
    def tag_str(self) -> str:
        if self.grandfathered_tag:
            return self.grandfathered_tag
        else:
            subtags: list[Any] = list(
                filter(
                    None,
                    [
                        self.lang,
                        self.extlang,
                        self.script,
                        self.region,
                        *self.variants.order_by("tags_through__index"),
                        *self.extensions.order_by("index"),
                    ],
                )
            )
        if self.private:
            subtags.extend(["x", self.private])

        return "-".join(map(str, subtags))

    @cached_property
    def description(self) -> str:
        """English description of what the tag represents."""
        if self.iana:
            return self.iana.first_description()
        if self.lang is None:
            if self.private is not None:
                return f'Private tag x-"{self.private}"'
            else:
                raise ValueError(
                    f"IANA record for grandfathered tag '{self}' not found"
                )
        parts = [self.lang.iana.first_description()]
        if self.extlang:
            parts.append("-")
            parts.append(self.extlang.iana.first_description())
        if self.region:
            parts.append("as used in")
            parts.append(self.region.iana.first_description())
        if self.script and self.script.script != self.lang.suppress_script:
            parts.append("as written in")
            parts.append(self.script.iana.first_description())
        for variant in self.variants_through.order_by("index"):
            parts.append("-")
            parts.append(variant.variant.iana.first_description())
        parts.extend(map(str, self.extensions.order_by("index")))
        return " ".join(parts)

    def lite(self) -> jangle.lite.LanguageTag:
        lite = jangle.lite.LanguageTag(
            grandfathered=self.grandfathered_tag,
            langtag=None,
            private=None,
        )
        if self.is_private:
            lite.private = self.private
        elif self.lang:
            langtag = jangle.lite.LangTag(self.lang.code, private=self.private)
            if self.extlang:
                langtag.extlang = self.extlang.code
            if self.script:
                langtag.script = self.script.script.code
            if self.region:
                langtag.region = self.region.code
            langtag.variants = [
                variant_through.variant.text
                for variant_through in self.variants_through.order_by("index")
            ]
            langtag.extensions = [
                jangle.lite.Extension(
                    singleton=extension.singleton,
                    texts=[
                        text.text for text in extension.texts.order_by("index")
                    ],
                )
                for extension in self.extensions.order_by("index")
            ]
            lite.langtag = langtag
        return lite

    def __str__(self) -> str:
        return self.tag_str

    objects = LanguageTagManager()

    class Meta:
        constraints = [
            models.CheckConstraint(
                check=models.Q(lang__isnull=False)
                | models.Q(grandfathered_tag__isnull=False)
                | models.Q(private__isnull=False),
                name="langtag_or_grandfathered_or_private",
            )
        ]


class LanguageTagVariantSubtag(models.Model):
    """"""

    tag = models.ForeignKey(
        LanguageTag,
        related_name="variants_through",
        on_delete=models.CASCADE,
    )
    """"""
    variant = models.ForeignKey(
        VariantSubtag,
        related_name="tags_through",
        on_delete=models.CASCADE,
    )
    """"""
    index = models.PositiveSmallIntegerField(default=0)
    """"""

    class Meta:
        unique_together = (("tag", "index"), ("tag", "variant"))
        ordering = ["tag", "index"]


class ExtensionSubtag(models.Model):
    """"""

    texts: "models.manager.RelatedManager[ExtensionSubtagText]"
    """"""
    tag = models.ForeignKey(
        LanguageTag,
        related_name="extensions",
        on_delete=models.CASCADE,
    )
    """"""
    singleton = models.CharField(max_length=1)
    """"""
    index = models.PositiveSmallIntegerField(default=0)
    """"""

    def __str__(self) -> str:
        texts = self.texts.order_by("index")
        return "-".join([self.singleton, *map(str, texts)])

    class Meta:
        unique_together = (("tag", "index"), ("tag", "singleton"))
        ordering = ["tag", "index"]


class ExtensionSubtagText(models.Model):
    """"""

    extension = models.ForeignKey(
        ExtensionSubtag,
        related_name="texts",
        on_delete=models.CASCADE,
    )
    """"""
    text = models.CharField(max_length=8)
    """"""
    index = models.PositiveSmallIntegerField(default=0)
    """"""

    def __str__(self) -> str:
        return self.text

    class Meta:
        unique_together = ("extension", "index")
        ordering = ["extension", "index"]



from __future__ import annotations

import csv
import warnings
from typing import Optional
from zipfile import ZipFile

import requests
from django.db import models
from requests.compat import urljoin

from jangle.readers import SilTableReader
from jangle.utils import BatchedCreateManager


class InvalidISO639Error(ValueError):
    pass


class ISOLanguageCodesManager(BatchedCreateManager["ISOLanguageCodes"]):
    def get_from_ietf(self, code: str) -> ISOLanguageCodes:
        if len(code) == 2:
            return self.get(part_1=code)
        elif len(code) == 3:
            for field in ["part_2b", "part_2t"]:
                try:
                    return self.get(**{field: code})
                except self.model.DoesNotExist:
                    continue
        else:
            raise InvalidISO639Error(code)
        raise self.model.DoesNotExist(code)

    def register(self, clear=True, batch_size=64) -> None:
        """Saves ISO 639-2 and 639-1 codes from the Library of Congress.
        See https://www.loc.gov/standards/iso639-2/ISO-639-2_utf-8.txt.
        """
        response = requests.get(
            "https://www.loc.gov/standards/iso639-2/ISO-639-2_utf-8.txt"
        )
        response.raise_for_status()
        response.encoding = "utf-8"
        if clear:
            self.all().delete()
        fieldnames = [
            "part_2b",
            "part_2t",
            "part_1",
            "names_en",
            "names_fr",
        ]
        self.batched_create(
            (
                self.model(
                    part_2b=row["part_2b"][-3:],
                    part_2t=row["part_2t"] or row["part_2b"][-3:],
                    part_1=row["part_1"] or None,
                    names_en=row["names_en"],
                    names_fr=row["names_fr"],
                )
                for row in csv.DictReader(
                    response.iter_lines(decode_unicode=True),
                    fieldnames,
                    delimiter="|",
                )
            ),
            batch_size,
        )


class ISOLanguageCodes(models.Model):
    """ISO 639-2 and 639-1 language codeset,
    saved from the Library of Congress.
    """

    language: Optional[ISOLanguage]
    """"""
    part_2b = models.CharField(
        "bibliographic code (ISO 639-2/B)",
        unique=True,
        max_length=3,
    )
    """Bibliographic code (ISO 639-2/B)."""
    part_2t = models.CharField(
        "terminological code (ISO 639-2/T)",
        unique=True,
        max_length=3,
    )
    """Terminological code (ISO 639-2/T)."""
    part_1 = models.CharField(
        "alpha-2 code (ISO 639-1)",
        null=True,
        max_length=2,
    )
    """Alpha-2 code (ISO 639-1)."""
    names_en = models.CharField(
        "noms français",
        unique=True,
        max_length=75,
    )
    """English names.
    Separated with ";" if multiple exist.
    """
    names_fr = models.CharField(
        "English names",
        unique=True,
        max_length=75,
    )
    """French names.
    Separated with ";" if multiple exist.
    """

    @property
    def ietf(self) -> str:
        """Code used in IETF language tags."""
        return self.part_1 or self.part_2b

    def __str__(self) -> str:
        return self.ietf

    objects = ISOLanguageCodesManager()

    class Meta:
        verbose_name = "ISO 639-2 and 639-1 language codeset"


class SimpleISOLanguageCollectionManager(
    BatchedCreateManager["SimpleISOLanguageCollection"]
):
    def get_from_ietf(self, code: str) -> SimpleISOLanguageCollection:
        return self.get(part_5=code)

    def register(self, clear=True, batch_size=64) -> None:
        """Saves basic ISO 639-5 data from the Library of Congress.
        See http://id.loc.gov/vocabulary/iso639-5.tsv.
        """
        response = requests.get("http://id.loc.gov/vocabulary/iso639-5.tsv")
        response.raise_for_status()

        if clear:
            self.all().delete()
        self.batched_create(
            (
                self.model(
                    part_5=row["code"],
                    names_en=row["Label (English)"],
                    names_fr=row["Label (French)"],
                )
                for row in csv.DictReader(
                    response.iter_lines(decode_unicode=True),
                    dialect="excel-tab",
                )
            ),
            batch_size,
        )


class SimpleISOLanguageCollection(models.Model):
    """Basic data for an ISO 639-5 language collection."""

    part_5 = models.CharField(
        "ISO 639-5 alpha-3 code", unique=True, max_length=3
    )
    """ISO 639-5 alpha-3 code."""
    names_en = models.CharField(
        "English names",
        unique=True,
        max_length=75,
    )
    """English names.
    Separated with ";" if multiple exist.
    """
    names_fr = models.CharField(
        "noms français",
        unique=True,
        max_length=75,
    )
    """French names.
    Separated with ";" if multiple exist.
    """

    @property
    def loc_uri(self) -> str:
        """URI on the Library of Congress.
        Contains MADS/SKOS RDF data.
        """
        return urljoin("http://id.loc.gov/vocabulary/iso639-5", self.part_5)

    @property
    def ietf(self) -> str:
        """Code used in IETF language tags."""
        return self.part_5

    def __str__(self) -> str:
        return self.part_5

    objects = SimpleISOLanguageCollectionManager()

    class Meta:
        verbose_name = "simple ISO 639-5 language collection"


class ISOLanguageQuerySet(models.QuerySet["ISOLanguage"]):
    def macrolanguages(self) -> ISOLanguageQuerySet:
        return self.filter(scope=ISOLanguage.Scope.MACROLANGUAGE)

    def individuals(self) -> ISOLanguageQuerySet:
        return self.filter(scope=ISOLanguage.Scope.INDIVIDUAL)

    def specials(self) -> ISOLanguageQuerySet:
        return self.filter(scope=ISOLanguage.Scope.SPECIAL)

    def get_from_ietf(self, code: str) -> ISOLanguage:
        if len(code) == 2:
            return self.get(codes__part_1=code)
        elif len(code) == 3:
            for field in ["part_3", "codes__part_2b", "codes__part_2t"]:
                try:
                    return self.get(**{field: code})
                except self.model.DoesNotExist:
                    continue
        else:
            raise InvalidISO639Error(code)
        raise self.model.DoesNotExist(code)


class ISOLanguageManager(BatchedCreateManager["ISOLanguage"]):
    def get_queryset(self) -> ISOLanguageQuerySet:
        return ISOLanguageQuerySet(self.model, using=self._db)

    def individuals(self) -> ISOLanguageQuerySet:
        return self.get_queryset().individuals()

    def macrolanguages(self) -> ISOLanguageQuerySet:
        return self.get_queryset().macrolanguages()

    def specials(self) -> ISOLanguageQuerySet:
        return self.get_queryset().specials()

    def get_from_ietf(self, code: str) -> ISOLanguage:
        return self.get_queryset().get_from_ietf(code)

    def register(
        self, clear=True, batch_size=64, zf: Optional[ZipFile] = None
    ) -> None:
        """Saves ISO 639-3 languages from SIL International."""

        def generate():
            with SilTableReader("iso-639-3", zf) as reader:
                for row in reader:
                    codes = None
                    if row["Part2B"]:
                        try:
                            codes = ISOLanguageCodes.objects.get(
                                part_2b=row["Part2B"],
                                part_2t=row["Part2T"] or None,
                                part_1=row["Part1"] or None,
                            )
                        except ISOLanguageCodes.DoesNotExist:
                            warnings.warn(str(row))
                    yield self.model(
                        codes=codes,
                        part_3=row["Id"],
                        scope=row["Scope"],
                        lang_type=row["Language_Type"],
                        ref_name=row["Ref_Name"],
                        comment=row["Comment"] or None,
                    )

        if clear:
            self.all().delete()
        self.batched_create(generate(), batch_size)

        # Macrolanguage mappings
        mappings: dict[str, list[str]] = dict()
        last_m_id = ""
        with SilTableReader("iso-639-3-macrolanguages", zf) as reader:
            for row in reader:
                if row["M_Id"] != last_m_id:
                    last_m_id = row["M_Id"]
                    mappings[last_m_id] = []
                if row["I_Status"] == "A":  # Active
                    mappings[last_m_id].append(row["I_Id"])
                for m_part_3, i_part_3s in mappings.items():
                    self.filter(part_3__in=i_part_3s).update(
                        macrolanguage=self.get(part_3=m_part_3)
                    )


class ISOLanguage(models.Model):
    """Represents an ISO 639-3 language,
    saved from SIL international"""

    class LanguageType(models.TextChoices):
        ANCIENT = "A", "ancient"
        CONSTRUCTED = "C", "constructed"
        EXTINCT = "E", "extinct"
        HISTORICAL = "H", "historical"
        LIVING = "L", "living"
        SPECIAL = "S", "special"

    class Scope(models.TextChoices):
        INDIVIDUAL = "I", "individual"
        MACROLANGUAGE = "M", "macrolanguage"
        SPECIAL = "S", "special"

    names: "models.manager.RelatedManager[ISOLanguageName]"
    """"""
    codes = models.OneToOneField(
        ISOLanguageCodes,
        related_name="language",
        null=True,
        unique=True,
        on_delete=models.CASCADE,
    )
    """ISO 693-1 and 639-2 codes."""
    ref_name = models.CharField(
        "reference name",
        unique=True,
        max_length=150,
    )
    """Reference name."""
    part_3 = models.CharField(
        "alpha-3 code (ISO 639-3)",
        unique=True,
        max_length=3,
    )
    """Alpha-3 code (ISO 639-3)."""
    lang_type = models.CharField(
        "type",
        choices=LanguageType.choices,
        max_length=1,
    )
    """Type."""
    scope = models.CharField(choices=Scope.choices, max_length=1)
    """"""
    comment = models.CharField(null=True, max_length=150)
    """"""
    macrolanguage = models.ForeignKey(
        "self",
        related_name="individuals",
        null=True,
        on_delete=models.SET_NULL,
    )
    """Macrolanguage."""

    @property
    def ietf(self) -> str:
        """Shortest ISO 639 code (part 1 or 3)."""
        return self.codes and self.codes.part_1 or self.part_3

    objects = ISOLanguageManager()

    def __str__(self) -> str:
        return f"{self.part_3} - {self.ref_name}"

    class Meta:
        verbose_name = "ISO 639-3 language"


class ISOLanguageNameManager(BatchedCreateManager["ISOLanguageName"]):
    def register(
        self, clear=True, batch_size=64, zf: Optional[ZipFile] = None
    ) -> None:
        """Saves ISO 639-3 language names from SIL International."""
        if clear:
            self.all().delete()

        with SilTableReader("iso-639-3_Name_Index", zf) as reader:
            self.batched_create(
                (
                    self.model(
                        iso_lang=ISOLanguage.objects.get(part_3=row["Id"]),
                        printable=row["Print_Name"],
                        inverted=row["Inverted_Name"],
                    )
                    for row in reader
                ),
                batch_size,
            )


class ISOLanguageName(models.Model):
    """Represents an English name for an ISO 639-3 language,
    saved from SIL International.
    """

    iso_lang = models.ForeignKey(
        ISOLanguage,
        related_name="names",
        on_delete=models.CASCADE,
    )
    """ISO Language."""
    printable = models.CharField(
        "printable translated name",
        max_length=75,
    )
    """Printable translated name."""
    inverted = models.CharField(
        "inverted translated name",
        max_length=75,
    )
    """Inverted translated name."""

    def __str__(self) -> str:
        return self.printable

    objects = ISOLanguageNameManager()

    class Meta:
        verbose_name = "ISO 639-3 language name"



import re
from itertools import islice
from typing import Any, Iterable, Optional, TypeVar

from django.db import models


class StrReprCls:
    """Provides __repr__ showing
    the object's module, class name and string representation.
    """

    def __repr__(self) -> str:
        return f"<{self.__module__}.{type(self).__qualname__} '{str(self)}'>"


def findattr(obj, path) -> Optional[Any]:
    """This thing is goofy. Don't use it."""
    if isinstance(path, tuple):
        for subpath in path:
            attr = findattr(obj, subpath)
            if attr is not None:
                return attr
    elif isinstance(path, list):
        if len(path) < 1:
            raise ValueError("empty path")
        obj = getattr(obj, path.pop(0))
        if not path:
            return obj
        elif len(path) > 1:
            return findattr(obj, path)
    else:
        raise TypeError(
            "path should be list[str | tuple] or tuple[str | list]"
        )


def choice_from_iana(choices: type[models.TextChoices], string: str) -> Any:
    """Chooses a Django TextChoices field from text used in the IANA registry."""
    return getattr(choices, string.upper().replace("-", "_"))


_T = TypeVar("_T", bound=models.Model)


class BatchedCreateManager(models.Manager[_T]):
    """"""

    def batched_create(self, objs: Iterable, batch_size=64, **kwargs) -> None:
        """Bulk-creates objects in batches of specified size."""
        while True:
            batch = list(islice(objs, batch_size))
            if not batch:
                break
            self.bulk_create(batch, batch_size, **kwargs)


def split_subtags(string: str) -> list[str]:
    """Splits a tag or string of subtags."""
    return re.split(r"(?<!(?<![^\-\s])[^\-\s])[\-\s]", string.strip("-"))



from django.db import models

from .languages import ISOLanguage


class Region(models.Model):
    names: "models.manager.RelatedManager[RegionName]"
    """"""
    class Tier(models.IntegerChoices):
        GLOBAL = 0, "global"
        REGIONAL = 1, "regional"
        CONTINENTAL = 2, "continental"
        SUB_REGIONAL = 3, "sub-regional"
        INTERMEDIARY = 4, "intermediary"
        COUNTRY = 5, "country or area"

    tier = models.PositiveSmallIntegerField(choices=Tier.choices)
    """Classification."""
    no = models.PositiveSmallIntegerField(
        "ISO 3166-1 numeric / UN M.49 code",
        unique=True,
    )
    """ISO 3166-1 numeric / UN M.49 code."""
    parent = models.ForeignKey(
        "self",
        null=True,
        on_delete=models.CASCADE,
    )
    """Parent region encompassing this one, if existing."""

    @property
    def no_str(self) -> str:
        """Formatted string of numeric code."""
        return "{:03d}".format(self.no)

    def get_name(self, iso_language: ISOLanguage) -> str:
        return self.names.get(iso_language=iso_language).name

    def __str__(self) -> str:
        return self.get_name(ISOLanguage.objects.get_from_ietf("en"))


class RegionName(models.Model):
    region = models.ForeignKey(
        Region,
        related_name="names",
        on_delete=models.CASCADE,
    )
    """"""
    name = models.TextField(max_length=150)
    """"""
    iso_lang = models.ForeignKey(
        ISOLanguage,
        on_delete=models.PROTECT,
    )
    """ISO language."""

    class Meta:
        unique_together = ("region", "iso_lang")


class ISORegion(models.Model):
    """ISO 3166 region."""

    region = models.OneToOneField(
        Region,
        related_name="iso",
        on_delete=models.CASCADE,
    )
    """"""
    alpha_2 = models.CharField(
        "ISO 3166-1 alpha-2 code",
        unique=True,
        max_length=2,
    )
    """ISO 3166-1 alpha-2 code."""
    alpha_3 = models.CharField(
        "ISO 3166-1 alpha-3 code",
        unique=True,
        max_length=3,
    )
    """ISO 3166-1 alpha-3 code."""
    in_ldc = models.BooleanField(
        "Least Developed Countries",
        default=False,
    )
    """In UN Least Developed Countries (LDC)."""
    in_lldc = models.BooleanField(
        "Land Locked Developing Countries",
        default=False,
    )
    """In UN Land Locked Developing Countries (LLDC)."""
    in_sids = models.BooleanField(
        "Small Island Developing States",
        default=False,
    )
    """In UN Small Island Developing States (SIDS)."""

    def __str__(self) -> str:
        return str(self.region)

    class Meta:
        verbose_name = "ISO 3166 Region"



# Generated by Django 4.1.3 on 2022-11-13 05:27

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("jangle", "0001_initial"),
    ]

    operations = [
        migrations.RenameField(
            model_name="regionsubtag",
            old_name="ext_data",
            new_name="region",
        ),
    ]



from pathlib import Path

import django
from django.conf import settings

BASE_DIR = Path(__file__).resolve().parents[1]


def boot_django() -> None:
    """Boots Django with jangle and a sqlite DB for development."""
    settings.configure(
        BASE_DIR=BASE_DIR,
        DEBUG=True,
        DATABASES={
            "default": {
                "ENGINE": "django.db.backends.sqlite3",
                "NAME": BASE_DIR.parent / "db.sqlite3",
            },
        },
        INSTALLED_APPS=[
            "jangle",
        ],
        LANGUAGE_CODE="en-US",
        TIME_ZONE="UTC",
        USE_TZ=True,
    )

    django.setup()



import sys
from django.core.management import execute_from_command_line
from jangle.boot_django import boot_django


def main() -> None:
    boot_django()
    execute_from_command_line(sys.argv)


if __name__ == "__main__":
    main()



from django.core.management.base import BaseCommand, CommandParser

import io
import warnings
import zipfile

import requests
from requests.compat import urljoin

from jangle.models import (
    IANASubtagRegistry,
    ISOLanguage,
    ISOLanguageCodes,
    ISOLanguageName,
    LanguageTag,
    Script,
)
from jangle.readers import SIL_ISO_639_DOWNLOADS_URL, SIL_ISO_639_ZIPFILE


class Command(BaseCommand):
    help = "Saves Jangle data from various sources to the database."

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            "-c",
            "--clear",
            action="store_true",
            help="Deletes existing data",
        )
        parser.add_argument(
            "-C",
            "--clear-tags",
            action="store_true",
            help="Deletes existing data and language tags",
        )
        parser.add_argument(
            "-b",
            "--batch-size",
            default=128,
            help="Insert batch size",
        )

    def handle(self, *args, **options) -> None:
        if options["clear_tags"]:
            LanguageTag.objects.all().delete()
        clear = options["clear"] or options["clear_tags"]
        batch_size = options["batch_size"]
        self.stdout.write("Registering ISO 639-2 and 639-1 codes...")
        ISOLanguageCodes.objects.register(clear, batch_size)
        self.stdout.write("Requesting ISO 639-3 code tables as zip...")
        r = requests.get(
            urljoin(
                SIL_ISO_639_DOWNLOADS_URL,
                SIL_ISO_639_ZIPFILE + ".zip",
            ),
            stream=True,
        )
        try:
            r.raise_for_status()
            zf = zipfile.ZipFile(io.BytesIO(r.content))
        except requests.HTTPError as err:
            warnings.warn(err.args[0])
            self.stdout.write(
                self.style.WARNING(
                    "Could not retrieve ISO 639-3 zip, "
                    "requesting tables individually...",
                )
            )
            zf = None
        self.stdout.write("Saving ISO 639-3 language codes...")
        ISOLanguage.objects.register(clear, batch_size, zf)
        self.stdout.write("Saving ISO 639-3 language names...")
        ISOLanguageName.objects.register(clear, batch_size, zf)
        self.stdout.write("Registering ISO 15924 scripts...")
        Script.objects.register(clear, batch_size)
        self.stdout.write("Registering subtags and tags from IANA...")
        IANASubtagRegistry.objects.register(clear, batch_size)
        self.stdout.write(self.style.SUCCESS("Done!"))



import csv

import requests
from django.db import models

from jangle.utils import BatchedCreateManager


class ScriptManager(BatchedCreateManager["Script"]):
    def register(self, clear=True, batch_size=64) -> None:
        """Saves IS0 15924 scripts from unicode.org to the database.
        See https://www.unicode.org/iso15924/iso15924.txt.
        """
        r = requests.get("https://www.unicode.org/iso15924/iso15924.txt")
        r.raise_for_status()

        def line_is_valid(line: str) -> bool:
            line = line.strip()
            return bool(line and not line.startswith("#"))

        if clear:
            self.all().delete()
        fieldnames = [
            "code",
            "no",
            "names_en",
            "names_fr",
            "pva",
            "unicode_version",
            "script_date",
        ]
        self.batched_create(
            (
                self.model(
                    **{key: val or None for key, val in row.items()},
                )
                for row in csv.DictReader(
                    filter(line_is_valid, r.iter_lines(decode_unicode=True)),
                    fieldnames,
                    delimiter=";",
                )
            ),
            batch_size,
        )


class Script(models.Model):
    """Represents an ISO 15924 script,
    saved from https://www.unicode.org/iso15924/.
    """
    code = models.CharField("ISO 15924 code", unique=True, max_length=4)
    """ISO 15924 code."""
    no = models.PositiveSmallIntegerField("ISO 15924 number", unique=True)
    """ISO 15924 number."""
    names_en = models.CharField("English names", unique=True, max_length=75)
    """English names."""
    names_fr = models.CharField("noms français", unique=True, max_length=75)
    """French names."""
    pva = models.CharField("property value alias", null=True, max_length=150)
    """Unicode property value alias.
    See https://www.unicode.org/Public/UCD/latest/ucd/PropertyValueAliases.txt.
    """
    unicode_version = models.CharField(null=True, max_length=12)
    """"""
    script_date = models.DateField("date")
    """Date."""

    @property
    def no_str(self) -> str:
        """Formatted string of ISO 15924 number."""
        return "{:03d}".format(self.no)

    def __str__(self) -> str:
        return self.code

    objects = ScriptManager()

    class Meta:
        verbose_name = "ISO 15924 script"
